#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-20
@file:      test_pool.py
@contact:   georgewang1994@163.com
@desc:      测试协程池
"""

# import asyncio
import time
#
# from ammonia.backends.worker.pool import AsyncPool


def result_executor(i):
    print("%s: 等待1s" % i)
    time.sleep(1)
    return i


def success_executor(i):
    print("success", i)


def fail_executor(i):
    print("fail", i)

#
# async def run():
#     async with AsyncPool(worker_count=10) as pool:
#         for i in range(50):
#             await pool.apply_async(result_executor, success_executor, fail_executor, i)
#
#
# result_queue = asyncio.Queue()
# loop = asyncio.get_event_loop()
# loop.run_until_complete(run())


import asyncio

from ammonia.worker.pool import AsyncPool


async def run():
    async with AsyncPool(10, loop) as pool:
        for i in range(50):
            await pool.apply_async(result_executor, success_executor, fail_executor, i)

loop = asyncio.get_event_loop()

loop.run_until_complete(run())



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-25
@file:      listener.py
@contact:   georgewang1994@163.com
@desc:      起线程负责监听消息
"""

import logging
import threading
from queue import Empty

from kombu.mixins import ConsumerMixin

from ammonia.mq import TaskConnection, task_queues, TaskConsumer

logger = logging.getLogger(__name__)


class TaskConsumerWorker(ConsumerMixin):
    def __init__(self, connection, on_task_message):
        self.connection = connection
        self.on_task_message = on_task_message

    def get_consumers(self, Consumer, channel):
        return [TaskConsumer(channel=channel, queues=task_queues, prefetch_count=1, callbacks=[self.on_task_message],
                             accept=['pickle'])]


class TaskListener(object):
    """
    负责监听consumer，将consumer中的消息给取出来
    """
    def __init__(self, ready_queue, *args, **kwargs):
        self.task_consumer = None
        self.ready_queue = ready_queue
        self._connection = None

    def establish_connection(self):
        from ammonia.app import Ammonia
        if self.task_consumer:
            return

        self._connection = TaskConnection(hostname=Ammonia.conf["TASK_URL"],
                                          connect_timeout=Ammonia.conf["BACKEND_CONNECTION_TIMEOUT"])
        self._connection.connect()
        # prefetch_count 保证每次consumer在同一时间只能获取一个消息
        self.task_consumer = TaskConsumerWorker(self._connection, self.handle_task_message)

    def close_connection(self):
        if not self._connection:
            return

        self.task_consumer.should_stop = True
        self.task_consumer = None
        self._connection.close()
        self._connection = None

    def start(self):
        self.close_connection()
        self.establish_connection()
        self.consume()

    def stop(self):
        self.close_connection()

    def consume(self):
        """
        消费消息
        :return:
        """
        try:
            print("TaskListener: 获取消息中...")
            self.task_consumer.run()
        except KeyboardInterrupt:
            print("TaskListener: bye bye")
            pass

    def handle_task_message(self, body, message):
        print("TaskListener: 获取到消息%s" % body)
        self.ready_queue.put(body)
        message.ack()


class TaskQueueListener(threading.Thread):
    """
    负责监听ready_queue，将队列中的消息给取出来，加入到协程池
    """
    def __init__(self, ready_queue, process_callback, loop, *args, **kwargs):
        super(TaskQueueListener, self).__init__(name="task_queue_listener", target=self.consume, *args, **kwargs)
        self.daemon = True
        self.ready_queue = ready_queue
        self.process_callback = process_callback
        self.loop = loop

    def consume(self):
        """
        消费消息
        :return:
        """
        print("task queue listener start...")
        logging.info("task queue listener start...")
        while True:
            try:
                task_msg = self.ready_queue.get(timeout=1)
                if task_msg:
                    print("TaskQueueListener: 获取到消息%s" % task_msg)
                    task_name = task_msg["task_name"]
                    args, kwargs = task_msg["execute_args"], task_msg["execute_kwargs"]
                    self.ready_queue.task_done()
                    if not task_name:
                        print("task is None, stop running...")
                        continue

                    print("获取到消息中的 args: %s, kwargs: %s" % (args, kwargs))
                    self.loop.run_until_complete(self.process_callback(task_name, *args, **kwargs))
            except Empty:
                print("TaskQueueListener: 等待消息中...")
            except KeyboardInterrupt:
                print("TaskQueueListener: bye bye")

    def stop(self):
        self.join()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-24
@file:      mq.py
@contact:   georgewang1994@163.com
@desc:      ...
"""

from kombu import Consumer, Producer, Connection, Exchange, Queue

from ammonia import settings


# ---------------------------------- task mq ---------------------------------- #


class TaskConnection(Connection):
    hostname = settings.TASK_URL


class TaskExchange(Exchange):
    def __init__(self, name=None, channel=None, *args, **kwargs):
        super(TaskExchange, self).__init__(name=name, channel=channel, *args, **kwargs)


class TaskQueue(Queue):
    def __init__(self, name=None, routing_key=None, exchange=None, channel=None, *args, **kwargs):
        super(TaskQueue, self).__init__(
            name=name, exchange=exchange, routing_key=routing_key,
            channel=channel, *args, **kwargs
        )


task_exchange = TaskExchange(name=settings.TASK_EXCHANGE_NAME)
task_queues = [
    TaskQueue(name=settings.HIGH_TASK_QUEUE_NAME, routing_key=settings.HIGH_TASK_ROUTING_KEY, exchange=task_exchange),
    TaskQueue(name=settings.MID_TASK_QUEUE_NAME, routing_key=settings.MID_TASK_ROUTING_KEY, exchange=task_exchange),
    TaskQueue(name=settings.LOW_TASK_QUEUE_NAME, routing_key=settings.LOW_TASK_ROUTING_KEY, exchange=task_exchange),
]


class TaskConsumer(Consumer):
    def __init__(self, channel=None, queues=None, *args, **kwargs):
        super(TaskConsumer, self).__init__(channel=channel, queues=queues, *args, **kwargs)


class TaskProducer(Producer):
    def __init__(self, routing_key='', channel=None, exchange=None, *args, **kwargs):
        super(TaskProducer, self).__init__(routing_key=routing_key, channel=channel,
                                           exchange=exchange, *args, **kwargs)

    def publish_task(self, message, routing_key, exchange, declare):
        super(TaskProducer, self).publish(body=message, routing_key=routing_key, exchange=exchange, declare=declare,
                                          serializer='pickle')


# ---------------------------------- backend mq ---------------------------------- #


class BackendConnection(Connection):
    hostname = settings.BACKEND_URL


class BackendExchange(Exchange):
    def __init__(self, channel=None, *args, **kwargs):
        # 默认参数durable为True，auto_delete=False，保证持久化
        super(BackendExchange, self).__init__(channel=channel, *args, **kwargs)


class BackendQueue(Queue):
    def __init__(self, routing_key="", exchange=None, channel=None, *args, **kwargs):
        # 默认参数durable为True，auto_delete=False，保证持久化，并且用完即删除
        super(BackendQueue, self).__init__(
            exchang=exchange, routing_key=routing_key,
            channel=channel, *args, **kwargs
        )


class BackendConsumer(Consumer):
    def __init__(self, routing_key, channel=None, callbacks=None, *args, **kwargs):
        queues = [BackendQueue(routing_key=routing_key, channel=channel)]
        super(BackendConsumer, self).__init__(channel=channel, queues=queues, no_ack=False,
                                              callbacks=callbacks, *args, **kwargs)


class BackendProducer(Producer):
    def __init__(self, routing_key="", channel=None, exchange=None, *args, **kwargs):
        super(BackendProducer, self).__init__(routing_key=routing_key, channel=channel,
                                              exchange=exchange, *args, **kwargs)

    def publish_task(self, message):
        super(BackendProducer, self).publish(body=message, serializer="pickle")



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-11
@file:      exception.py
@contact:   georgewang1994@163.com
@desc:      错误
"""


class NotRegisterException(Exception):
    pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-11
@file:      test_backends.py
@contact:   georgewang1994@163.com
@desc:      测试数据库连接
"""

from unittest import TestCase

from ammonia.backends.backend import DbBackend, RedisBackend
from ammonia.backends.models import TaskStatusEnum
from ammonia.settings import TEST_CASE_BACKEND_URL
from ammonia.tests.test_base import TestDBBackendBase
from ammonia.utils import generate_random_uid


class TestDbBackend(TestDBBackendBase):
    """
    测试数据库持久化
    """
    def setUp(self):
        super(TestDbBackend, self).setUp()
        self.backend = DbBackend(TEST_CASE_BACKEND_URL)

    def test_insert_update_get_task(self):
        """
        测试插入更新任务
        """
        task_id1 = generate_random_uid()
        self.backend.mark_task_success(task_id=task_id1, result=100)
        result = self.backend.get_task(task_id1)
        self.assertEqual(result.status, TaskStatusEnum.SUCCESS.value)
        self.assertEqual(result.result, 100)

        task_id2 = generate_random_uid()
        self.backend.mark_task_fail(task_id=task_id2, result={"error": "error"})
        result = self.backend.get_task(task_id2)
        self.assertEqual(result.status, TaskStatusEnum.FAIL.value)
        self.assertEqual(result.traceback, {"error": "error"})


class TestRedisBackend(TestCase):
    """
    测试Redis持久化
    """
    def setUp(self):
        url = "redis://@localhost:6379/0"
        self.backend = RedisBackend(url)

    def test_mark_and_get_task(self):
        task_id1 = generate_random_uid()
        self.backend.mark_task_success(task_id=task_id1, result=3)
        result = self.backend.get_task_result(task_id1)
        self.assertEqual(result, 3)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-19
@file:      pool.py
@contact:   georgewang1994@163.com
@desc:      ...
"""

import asyncio
import functools
import logging


logger = logging.getLogger(__name__)


class AsyncPool(object):
    """
    协程池
    """
    def __init__(self, worker_count, loop=None):
        self.worker_count = worker_count
        self.loop = loop
        self.queue = asyncio.Queue(maxsize=self.worker_count)
        self._workers = []

    def start(self):
        logger.info("async pool start...")
        print("async pool start...")
        self._workers = [asyncio.ensure_future(self._worker(), loop=self.loop) for _ in range(self.worker_count)]

    async def __aenter__(self):
        self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.join()

    async def stop(self):
        await self.join()
        self.loop.stop()
        self.loop = None

    async def join(self):
        if not self._workers:
            return

        # 等待队列中的数据全部被处理完
        await self.queue.join()
        # 等待所有的worker被取消
        await asyncio.gather(*self._workers, loop=self.loop)
        self._workers = []

    async def _worker(self):
        while True:
            print("pool worker 等待获取消息...")
            item = await self.queue.get()

            future, executor_func, args, kwargs = item
            print("future:[%s], executor_func:[%s], args:[%s], kwargs:[%s]" % (future, executor_func, args, kwargs))
            try:
                result_value = executor_func(*args, **kwargs)
                print("获取到任务的结果:%s", result_value)
                future.set_result(result_value)
            except (KeyboardInterrupt, MemoryError, SystemExit) as e:
                future.set_exception(e)
                print("pool worker: bye bye")
            finally:
                self.queue.task_done()

    async def apply_async(self, executor_func, success_callback=None, error_callback=None, *args, **kwargs):
        future = asyncio.futures.Future(loop=self.loop)
        future.add_done_callback(functools.partial(self.on_receive, success_callback, error_callback))
        await self.queue.put((future, executor_func, args, kwargs))
        return future

    def on_receive(self, future, success_callback, error_callback):
        result_value = future.result()
        print("获取到任务的结果:%s", result_value)

        if success_callback:
            success_callback(result_value)

        if error_callback:
            error_callback(result_value)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-25
@file:      result.py
@contact:   georgewang1994@163.com
@desc:      任务结果封装
"""


class AsyncResult(object):
    def __init__(self, task_id, backend):
        self.task_id = task_id
        self.backend = backend

    def _get(self, timeout=None):
        """
        从backend中异步获取task_id对应的结果
        :return:
        """
        is_success, result = self.backend.get_task_result(self.task_id, timeout)
        if not is_success:
            return result

        return result

    def get(self, timeout=None):
        return self._get(timeout=timeout)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-06-15
@file:      config.py
@contact:   georgewang1994@163.com
@desc:      flask配置
"""

from enum import Enum

from ammonia.settings import BACKEND_URL, TEST_CASE_BACKEND_URL


class ConfigChoiceEnum(Enum):
    DEVELOPMENT = "dev"  # 线上开发环境
    TEST = "test"  # 测试环境
    TEST_CASE = "test_case"  # 测试用例环境


class _Config(object):
    # @see http://www.pythondoc.com/flask-sqlalchemy/config.html
    # 追踪对象的修改并且发送信号，默认开启
    SQLALCHEMY_TRACK_MODIFICATIONS = True

    @staticmethod
    def init_app(app):
        pass


class TestConfig(_Config):
    # 连接的数据库
    SQLALCHEMY_DATABASE_URI = BACKEND_URL

    # 显示数据库查询记录
    SQLALCHEMY_RECORD_QUERIES = True


class TestCaseConfig(_Config):
    # 连接的数据库
    SQLALCHEMY_DATABASE_URI = TEST_CASE_BACKEND_URL

    # 显示数据库查询记录
    SQLALCHEMY_RECORD_QUERIES = True


class DevConfig(_Config):
    # 连接的数据库
    SQLALCHEMY_DATABASE_URI = BACKEND_URL


_config = {
    "dev": DevConfig,
    "test": TestConfig,
    "test_case": TestCaseConfig,
}


def get_config(config):
    return _config.get(config, DevConfig)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-24
@file:      ammonia.py
@contact:   georgewang1994@163.com
@desc:      总控制
"""

from ammonia import settings
from ammonia.backends.backend import get_backend_by_settings
from ammonia.base.task import TaskManager

# 默认配置
DEFAULT_CONF = {
    "DEBUG": False,
    "TASK_URL": settings.get_task_url(),
    "BACKEND_URL": settings.get_backend_url(),
    "BACKEND_TYPE": "database",
    "TASK_CONNECTION_TIMEOUT": settings.TASK_CONNECTION_TIMEOUT,
    "BACKEND_CONNECTION_TIMEOUT": settings.BACKEND_CONNECTION_TIMEOUT,
}


class Ammonia(object):
    conf = DEFAULT_CONF
    backend = None

    def __init__(self, conf=None):
        if conf:
            self.conf.update(conf)

        backend_cls = get_backend_by_settings(self.conf["BACKEND_TYPE"])
        self.backend = backend_cls(self.conf["BACKEND_URL"])

    def task(self, *args, **kwargs):
        """
        task装饰器，创建task任务
        :param args:
        :param kwargs:
        :return:
        """
        def decorator(func):
            return TaskManager.create_task(func, self.backend, *args, **kwargs)
        return decorator

    def create_package(self, name, *args, **kwargs):
        """
        创建任务package
        :param name:
        :return:
        """
        return TaskManager.create_task_package(name, self.backend, *args, **kwargs)

    def update_conf(self, conf={}):
        return self.conf.update(conf)

    def get_conf(self):
        return self.conf

    def _check_conf(self):
        # todo: 检查配置是否正确
        pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-07-01
@file:      test_loader.py
@contact:   georgewang1994@163.com
@desc:      ...
"""

import os
import shutil
from unittest import TestCase

from ammonia.base.loader import Loader
from ammonia.settings import BASE_PROJECT_PATH


class TestLoader(TestCase):

    def setUp(self):
        self.package_name = "example_for_loader_test"
        self.example_path = os.path.join(BASE_PROJECT_PATH, self.package_name)
        if os.path.exists(self.example_path):
            shutil.rmtree(self.example_path)
        os.mkdir(self.example_path)

    def test_find_tasks(self):
        loader = Loader()
        mod_list = loader.find_tasks(self.package_name)
        self.assertEqual(len(mod_list), 0)

        with open(os.path.join(self.example_path, "__init__.py"), "w+") as p:
            pass
        with open(os.path.join(self.example_path, "first_tasks.py"), "w+") as fp:
            fp.write("""
def get_sum(a, b):
    return a + b""")

        mod_list = loader.find_tasks(self.package_name)
        self.assertEqual(len(mod_list), 1)
        self.assertEqual(mod_list[0].get_sum(1, 2), 3)

    def tearDown(self):
        shutil.rmtree(self.example_path)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-19
@file:      test_serializer.py
@contact:   georgewang1994@163.com
@desc:      测试序列化
"""
from unittest import TestCase

from ammonia.serializer import Serializer


class TestSerializer(TestCase):
    """
    测试序列化模块
    """
    def setUp(self):
        self.data = {"first": 1, "second": 2}

    def test_serializer(self):
        modules = ['pickle']
        for module in modules:
            serializer = Serializer(default_module=module)
            result = serializer.loads(serializer.dumps(self.data))
            self.assertDictEqual(result, self.data)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-12
@file:      utils.py
@contact:   georgewang1994@163.com
@desc:      基础函数
"""

import uuid


def generate_random_uid():
    return str(uuid.uuid1())



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-06-30
@file:      tasks.py
@contact:   georgewang1994@163.com
@desc:      ...
"""


from .ammonia import ammonia


@ammonia.task()
def get_sum(a, b):
    return a + b



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-06-29
@file:      start.py
@contact:   georgewang1994@163.com
@desc:      启动服务
"""

import sys

from ammonia.main import Worker, parse_options


def start():
    args = parse_options(sys.argv[1:])
    worker = Worker(worker_num=args.worker_num, project_name=args.project_name)
    worker.setup()
    worker.run()


if __name__ == "__main__":
    start()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-24
@file:      task.py
@contact:   georgewang1994@163.com
@desc:      任务相关的类
"""

import datetime
import random
import sys
from collections import Iterable

from ammonia import settings
from ammonia.base.registry import task_registry
from ammonia.base.result import AsyncResult
from ammonia.mq import TaskProducer, TaskConnection, task_exchange, task_queues
from ammonia.state import TaskStatusEnum
from ammonia.utils import generate_random_uid


class Crontab(object):
    def __init__(self, month="*", day="*", hour="*", minute="*", second="*"):
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second

    @property
    def data(self):
        return {
            "month": self.month,
            "day": self.day,
            "hour": self.hour,
            "minute": self.minute,
            "second": self.second,
        }


class BaseTask(object):
    def __init__(self, task_name, backend, *args, **kwargs):
        # 这里的args和kwargs是函数的，而装饰器的参数则是类的参数
        self.task_name = task_name
        self.task_id = ""
        self.status = TaskStatusEnum.CREATED.value
        self.routing_key = getattr(self, "routing_key", "")
        self.retry = getattr(self, "retry", 0)
        self.crontab = getattr(self, "crontab", None)
        self.eta = getattr(self, "eta", 0)  # 具体的执行时间，格式为datetime
        self.wait = getattr(self, "wait", 0)  # 等待执行的时间间隔，格式为delta
        self.execute_args = ()  # 执行函数的参数
        self.execute_kwargs = {}  # 执行函数的参数
        self.backend = backend
        self.result = None
        self.start_time = None

    def execute(self):
        raise NotImplemented

    @classmethod
    def get_task_producer(cls, channel, routing_key=None):
        return TaskProducer(channel=channel, routing_key=routing_key, exchange=task_exchange, serializer='json')

    def data(self):
        return {
            "task_name": self.task_name,
            "task_id": self.task_id,
            "status": self.status,
            "routing_key": self.routing_key,
            "retry": self.retry,
            "crontab": self.crontab.data if self.crontab else {},
            "eta": self.eta,
            "wait": self.wait,
            "execute_args": self.execute_args,
            "execute_kwargs": self.execute_kwargs,
        }

    @property
    def execute_time(self):
        return self.start_time

    def defer_async(self):
        """
        这里的参数是执行函数的参数，延迟执行
        :return:
        """
        from ammonia.app import Ammonia
        with TaskConnection(hostname=Ammonia.conf["TASK_URL"],
                            connect_timeout=Ammonia.conf["TASK_CONNECTION_TIMEOUT"]) as conn:
            # 如果路由不存在则随机路由
            if not self.routing_key:
                self.routing_key = random.choice(settings.TASK_ROUTING_KEY_LIST)
            print("发送消息给路由%s %s" % (self.routing_key, self.data()))
            producer = self.get_task_producer(channel=conn, routing_key=self.routing_key)
            producer.publish_task(self.data(), routing_key=self.routing_key,
                                  exchange=task_exchange, declare=task_queues)
            return AsyncResult(task_id=self.task_id, backend=self.backend)

    def __call__(self, *args, **kwargs):
        self.base_process_task(*args, **kwargs)
        return self.process_task(True)

    def defer(self, *args, **kwargs):
        self.base_process_task(*args, **kwargs)
        return self.process_task(False)

    def base_process_task(self, *args, **kwargs):
        self.start_time = datetime.datetime.now()
        if self.eta:
            self.start_time = self.eta
        elif self.wait:
            self.start_time = self.start_time + self.wait

        self.execute_args = args
        self.execute_kwargs = kwargs
        self.task_id = generate_random_uid()
        self.status = TaskStatusEnum.PREPARE.value

    def process_task(self, is_immediate=False):
        raise NotImplemented


class Task(BaseTask):

    def __str__(self):
        return "task[%s-%s]" % (self.task_name, self.status)

    def process_task(self, is_immediate=False):
        """
        :param is_immediate: 是否立即执行
        :return:
        """
        # 如果是直接调用，则直接计算返回
        if is_immediate:
            return task_trace_execute(self)

        return self.defer_async()


class TaskPackage(BaseTask):
    """
    将task都封装到package中，用于执行一系列任务
    """
    def __init__(self, task_name, backend, *args, **kwargs):
        super(TaskPackage, self).__init__(task_name, backend, *args, **kwargs)
        self.task_list = []
        self.is_dependent = getattr(self, "dependent", False)  # 是否前一个任务依赖于上一个任务的结果

    def register(self, task):
        if not task or task in self.task_list:
            return

        self.task_list.append(task)

    def __str__(self):
        return "package[%s-%s]" % (self.task_name, self.status)

    def process_task(self, is_immediate=False):
        """
        :param is_immediate: 是否立即执行
        :return:
        """
        if not self.task_list:
            raise Exception("任务包中的任务为空")

        if len(self.task_list) <= 1:
            raise Exception("任务包中任务个数至少为两个")

        task_id = generate_random_uid()
        self.status = TaskStatusEnum.PREPARE.value
        # 如果是直接调用，则直接计算返回
        if is_immediate:
            return package_trace_execute(self)

        return self.defer_async()


class TaskManager(object):
    base_task_class = Task
    base_package_class = TaskPackage

    def __init__(self, task_id, *args, **kwargs):
        self.task_id = task_id

    @classmethod
    def to_message(cls, task):
        return task.data()

    @classmethod
    def _get_task_execute_func(cls, execute_func):

        def execute(self, *args, **kwargs):
            return execute_func(*args, **kwargs)

        return execute

    @classmethod
    def create_task_class(cls, execute_func, **kwargs):

        task_execute_func = cls._get_task_execute_func(execute_func)

        task_settings = {
            'execute': task_execute_func,
            '__doc__': execute_func.__doc__,
            '__module': execute_func.__module__,
        }
        # 将执行函数的参数和装饰器的参数合并到一块
        task_settings.update(kwargs)

        return type(execute_func.__name__, (cls.base_task_class,), task_settings)

    @classmethod
    def create_task(cls, execute_func, backend, *args, **kwargs):
        task_module = sys.modules[execute_func.__module__]
        task_name = ".".join([task_module.__name__, execute_func.__name__, ])
        task_cls = cls.create_task_class(execute_func, **kwargs)
        task = task_cls(task_name, backend, *args, **kwargs)
        task_registry.register(task)

        # 如果填写了package的话，需要注册到package
        if kwargs.get("package"):
            if isinstance(kwargs["package"], str):
                package_name = kwargs["package"]
            elif isinstance(kwargs["package"], TaskPackage):
                package_name = kwargs["package"].task_name
            else:
                raise TypeError("package填写错误")

            task_package = task_registry.task(package_name)
            if task_package:
                task_package.register(task)
        return task

    @classmethod
    def create_task_package_class(cls, package_name, **kwargs):
        # 将执行函数的参数和装饰器的参数合并到一块
        return type(package_name, (cls.base_package_class,), kwargs)

    @classmethod
    def create_task_package(cls, package_name, backend, *args, **kwargs):
        task_package_cls = cls.create_task_package_class(package_name, **kwargs)
        task_package = task_package_cls(package_name, backend, *args, **kwargs)
        task_registry.register(task_package)
        return task_package

    @classmethod
    async def execute_task(cls, pool, task_name, *args, **kwargs):
        task = task_registry.task(task_name)
        if not task:
            print("task is None, stop running...")
            return

        print("开始处理任务 task_name: %s, task: %s, args: %s, kwargs: %s" % (task_name, task, args, kwargs))
        await pool.apply_async(task_trace_execute, cls.on_task_success, cls.on_task_fail, task, *args, **kwargs)

    @classmethod
    def on_task_success(cls, return_value):
        print("任务执行成功，返回结果为%s" % return_value)

    @classmethod
    def on_task_fail(cls, return_value):
        print("任务执行失败，返回结果为%s" % return_value)


class BaseTrace(object):
    task = None
    result = None

    def __init__(self, task):
        self.task_or_package = task

    def _execute(self, *args, **kwargs):
        raise NotImplemented

    def execute(self, *args, **kwargs):
        print("任务开始执行 task: %s, args: %s, kwargs: %s" % (self.task_or_package, args, kwargs))
        retry_num = self.task_or_package.retry
        result = None

        if retry_num:
            while retry_num:
                success, result = self._execute(*args, **kwargs)
                if success:
                    self.do_exec_success(result)
                    return result
                retry_num -= 1
        else:
            success, result = self._execute(*args, **kwargs)
            if success:
                self.do_exec_success(result)
                print("任务[%s]执行成功, 结果为[%s]" % (self.task, result))
                return result
            else:
                print("任务[%s]执行失败, 结果为[%s]" % (self.task, result))

        # 记录到数据库中
        if self.task_or_package.retry:
            self.do_exec_retry(result)
        else:
            self.do_exec_fail(result)
        return None

    def do_exec_func(self, *args, **kwargs):
        raise NotImplemented

    def do_exec_success(self, return_value):
        self.task_or_package.status = TaskStatusEnum.SUCCESS.value
        self.task_or_package.backend.mark_task_success(task_id=self.task.task_id, result=return_value)

    def do_exec_fail(self, return_value):
        self.task_or_package.status = TaskStatusEnum.FAIL.value
        self.task_or_package.backend.mark_task_fail(task_id=self.task.task_id, result=return_value)

    def do_exec_retry(self, return_value):
        self.task_or_package.status = TaskStatusEnum.RETRY.value
        self.task_or_package.backend.mark_task_retry(task_id=self.task.task_id, result=return_value)


class TaskTrace(BaseTrace):
    """
    一个task对应一个trace
    """
    def _execute(self, *args, **kwargs):
        try:
            result = self.do_exec_func(*args, **kwargs)
            return True, result
        except Exception as e:
            error = e.args[0]
            return False, error

    def do_exec_func(self, *args, **kwargs):
        self.task_or_package.status = TaskStatusEnum.PROCESS.value
        self.result = self.task_or_package.execute(*args, **kwargs)
        return self.result


class TaskPackageTrace(BaseTrace):
    """
    一个task对应一个trace
    """
    def _execute(self, *args):
        idx = 0
        last_result = None
        result = None
        while idx < len(self.task_or_package.task_list):
            task = self.task_or_package.task_list[idx]
            if not self.task_or_package.is_dependent:
                # 有些任务并不需要参数，因此填空
                sub_args = args[idx] if idx < len(args) else ()
            else:
                # 任务依赖遇上一个任务的结果作为参数
                if idx > 0:
                    sub_args = last_result if isinstance(last_result, Iterable) else (last_result,)
                else:
                    sub_args = args[0] if args else ()
            try:
                result = self.do_exec_func(task, *sub_args)
                last_result = result
                print("任务[%s]执行成功, 结果为[%s]" % (task, result))
                idx += 1
                continue
            except Exception as e:
                error = e.args[0]
                return False, error

        return True, result

    def do_exec_func(self, task, *args):
        self.task_or_package.status = TaskStatusEnum.PROCESS.value
        self.result = task.execute(*args)
        return self.result


def task_trace_execute(task):
    return TaskTrace(task=task).execute(*task.execute_args, **task.execute_kwargs)


def package_trace_execute(package):
    return TaskPackageTrace(task=package).execute(*package.execute_args)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-06-30
@file:      ammonia.py
@contact:   georgewang1994@163.com
@desc:      ...
"""

from ammonia.app import Ammonia


ammonia = Ammonia()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-07-09
@file:      clock.py
@contact:   georgewang1994@163.com
@desc:      定期处理任务
"""

import datetime
import threading
import time

from ammonia.base.registry import task_registry


class Schedule(threading.Thread):
    """
    每隔一段时间检查是否存在符合时间的任务执行
    """
    def __init__(self, ready_queue, *args, **kwargs):
        super(Schedule, self).__init__(*args, **kwargs)
        self.daemon = True
        self.ready_queue = ready_queue
        self.time_task_list = []  # 存储设置了时间的任务
        self.last_sync_time = None

    def sort_tasks(self):
        self.time_task_list.sort(key=lambda x: x.start_time)

    def get_need_execute_tasks(self, start_time):
        result = []
        next_internal_time = None

        # 计算需要执行的任务，并且将其删除，以及需要等待下一个任务的间隔时间
        for idx in range(len(self.time_task_list))[::-1]:
            task = self.time_task_list[idx]
            if start_time < task.start_time < start_time + datetime.timedelta(seconds=1):
                result.append(task)
                del self.time_task_list[idx]

            # 倒序后首次出现的下一个任务即需要执行的下一个任务
            if not next_internal_time:
                next_task = self.time_task_list[idx + 1] if idx + 1 < len(self.time_task_list) else None
                if next_task:
                    next_internal_time = (next_task.start_time - start_time).total_seconds()

        # 默认等待很长的时间
        if next_internal_time is None:
            next_internal_time = 1000

        return result, next_internal_time

    def run(self):
        print("schedule start running...")
        for task in task_registry.values():
            if task.eta or task.wait:
                self.time_task_list.append(task)

        print("schedule check [%s] execute tasks" % (len(self.time_task_list)))
        while True:
            self.sort_tasks()
            start_time = datetime.datetime.now()

            execute_tasks, next_internal_time = self.get_need_execute_tasks(start_time)
            print("schedule found [%s] tasks and wait [%s] seconds" % (len(execute_tasks), next_internal_time))
            for task in execute_tasks:
                self.ready_queue.put(task.data())  # todo: 可能时间不准，因为read_queue还是可能会有等待的时间

            print("schedule execute task from start_time:[%s] to end_time:[%s]" % (self.last_sync_time, start_time))
            self.last_sync_time = start_time
            time.sleep(next_internal_time)

    def stop(self):
        self.join()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-19
@file:      serializer.py
@contact:   georgewang1994@163.com
@desc:      序列化封装
"""
import zlib
from importlib import import_module


class Serializer(object):
    def __init__(self, default_module='pickle'):
        try:
            serializer_model = import_module(default_module)
            self._serializer = serializer_model
        except ImportError:
            raise ImportError(u"导入序列化模块失败")

    def dumps(self, data):
        return zlib.compress(self._serializer.dumps(data))

    def loads(self, data):
        return self._serializer.loads(zlib.decompress(data))


serializer = Serializer()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-24
@file:      registry.py
@contact:   georgewang1994@163.com
@desc:      注册任务
"""

from ammonia.exception import NotRegisterException


class Registry(dict):
    """
    任务注册
    """
    def _clean(self):
        self.clear()

    def register(self, task):
        """
        :param task:
        :return:
        """
        self[task.task_name] = task

    def unregister(self, task):
        """
        :param task:
        :return:
        """
        self.pop(task.task_name, None)

    def task(self, task_name):
        """

        :param task_name:
        :return:
        """
        try:
            return self[task_name]
        except NotRegisterException:
            raise NotRegisterException(task_name)


task_registry = Registry()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-12
@file:      settings.py
@contact:   georgewang1994@163.com
@desc:      配置文件
"""

import os


# 是否正在调试
DEBUG = False

# 存储的主机名
BACKEND_HOSTNAME = "localhost"

# 存储的端口号
BACKEND_PORT = "3306"

# 存储的用户名
BACKEND_USER = "root"

# 存储的密码
BACKEND_PASSWORD = "123456"

# 存储的数据库名字
BACKEND_NAME = "ammonia"

# 数据信息（数据库类型+数据库驱动名称）
BACKEND_TYPE = "mysql+pymysql"


# 连接超时时间
BACKEND_CONNECTION_TIMEOUT = 3


def get_backend_url(type=BACKEND_TYPE, user=BACKEND_USER, password=BACKEND_PASSWORD,
                    host=BACKEND_HOSTNAME, port=BACKEND_PORT, db_name=BACKEND_NAME):
    return f"{type}://{user}:{password}@{host}:{port}/{db_name}"


# 数据库URL(数据库类型+数据库驱动名称://用户名:口令@机器地址:端口号/数据库名)
BACKEND_URL = get_backend_url()

# 测试用例的数据库名字
TEST_CASE_BACKEND_NAME = f"test_{BACKEND_NAME}"

# 测试用例数据库URL
TEST_CASE_BACKEND_URL = get_backend_url(db_name=TEST_CASE_BACKEND_NAME)

# 连接超时时间
TASK_CONNECTION_TIMEOUT = 3


# 任务队列的主机名
TASK_HOSTNAME = "localhost"

# 任务队列的端口号
TASK_PORT = "5672"

# 任务队列的用户名
TASK_USER = "guest"

# 任务队列的密码
TASK_PASSWORD = "guest"

# 数据信息
TASK_TYPE = "amqp"


# 任务队列的地址
TASK_URL = f"{TASK_TYPE}://{TASK_USER}:{TASK_PASSWORD}@{TASK_HOSTNAME}:{TASK_PORT}//"


def get_task_url(type=TASK_TYPE, user=TASK_USER, password=TASK_PASSWORD,
                 host=TASK_HOSTNAME, port=TASK_PORT):
    return f"{type}://{user}:{password}@{host}:{port}//"


# 任务队列名
HIGH_TASK_QUEUE_NAME = "high_task"
MID_TASK_QUEUE_NAME = "mid_task"
LOW_TASK_QUEUE_NAME = "low_task"

# 任务队列路由key
HIGH_TASK_ROUTING_KEY = "high_queue"
MID_TASK_ROUTING_KEY = "mid_queue"
LOW_TASK_ROUTING_KEY = "low_queue"

TASK_ROUTING_KEY_LIST = [HIGH_TASK_ROUTING_KEY, MID_TASK_ROUTING_KEY, LOW_TASK_ROUTING_KEY]

# 任务交换器名字
TASK_EXCHANGE_NAME = "task"


# 项目路径
BASE_PROJECT_PATH = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))

# 获取任务模块的后缀名
TASK_MODULE_SUFFIX = "tasks"

# 是否是测试用例
IS_FOR_TEST_CASE = False



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-24
@file:      controller.py
@contact:   georgewang1994@163.com
@desc:      控制worker
"""

import asyncio
from queue import Queue

from ammonia.base.task import TaskManager
from ammonia.worker.listener import TaskListener, TaskQueueListener
from ammonia.worker.pool import AsyncPool
from ammonia.worker.schedule import Schedule


class WorkerController(object):
    """
    控制worker
    """
    def __init__(self, pool_worker_count=10):
        self.ready_queue = Queue()
        self.listener = TaskListener(ready_queue=self.ready_queue)
        self.schedule = Schedule(ready_queue=self.ready_queue)
        self.loop = asyncio.get_event_loop()
        self.queue_listener = TaskQueueListener(ready_queue=self.ready_queue,
                                                process_callback=self.process_task, loop=self.loop)
        self.pool = AsyncPool(pool_worker_count, self.loop)
        self.workers = (
            self.queue_listener,
            self.schedule,
            self.pool,
            self.listener,
        )

    async def process_task(self, task_name, *args, **kwargs):
        await TaskManager.execute_task(self.pool, task_name, *args, **kwargs)

    def start(self):
        """
        :return:
        """
        try:
            for worker in self.workers:
                worker.start()
        finally:
            self.close()

    def close(self):
        for worker in reversed(self.workers):
            if isinstance(worker, AsyncPool):
                # 等待stop函数完成
                loop = asyncio.get_event_loop()
                loop.run_until_complete(worker.stop())
            else:
                worker.stop()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-06-12
@file:      state.py
@contact:   georgewang1994@163.com
@desc:      状态机
"""

from enum import Enum, unique


@unique
class TaskStatusEnum(Enum):
    CREATED = "created"  # 初始创建
    PREPARE = "prepare"  # 准备处理
    PROCESS = "process"  # 正在处理
    SUCCESS = "success"  # 处理成功
    FAIL = "fail"        # 处理失败
    RETRY = "retry"        # 重试

    @classmethod
    def all_values(cls):
        return cls._value2member_map_.keys()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-11
@file:      backend.py
@contact:   georgewang1994@163.com
@desc:      RabbitMq的持久化，其是通过设置参数durable=True实现的
"""

import pickle
import time

from redis import Redis
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, exc

from ammonia import settings
from ammonia.backends.models import Task
from ammonia.mq import BackendConnection, BackendConsumer, BackendProducer
from ammonia.state import TaskStatusEnum

# 队列名字
QUEUE_NAME = "backend_queue"

# 交换器名字
EXCHANGE_NAME = "backend_exchange"


class BaseBackend(object):
    def __init__(self, backend_url, timeout=None):
        self.backend_url = backend_url
        self.timeout = timeout
        self._connection = None

    def mark_task_success(self, task_id, result=None):
        """
        标记任务为已经完成
        :param task_id:
        :param result:
        :return:
        """
        return NotImplemented

    def mark_task_retry(self, task_id, result=None):
        """
        标记任务为重试
        :param task_id:
        :param :
        :param result:
        :return:
        """
        return NotImplemented

    def mark_task_fail(self, task_id, result=None):
        """
        标记任务为失败
        :param task_id:
        :param result:
        :return:
        """
        return NotImplemented

    def get_task_result(self, task_id, timeout=None):
        """
        返回任务的结果
        :param task_id:
        :param timeout: 阻塞时间（None表示不阻塞直接拿，0表示无限阻塞，其余为阻塞秒数）
        :return:
        """
        return NotImplemented

    def establish_connection(self):
        """
        建立连接
        :return:
        """
        return NotImplemented

    def close_connection(self):
        """
        关闭连接
        :return:
        """
        return NotImplemented


class MQBackend(BaseBackend):
    """
    利用mq自身功能实现消息持久化
    """
    def __init__(self, backend_url=settings.BACKEND_URL, timeout=None):
        super(MQBackend, self).__init__(backend_url=backend_url, timeout=timeout)

    def establish_connection(self):
        if self._connection:
            return

        if not self._connection:
            self._connection = BackendConnection(hostname=self.backend_url,
                                                 connect_timeout=self.timeout)
            self._connection.connect()

        return self._connection

    def close_connection(self):
        if not self._connection:
            return

        if self._connection:
            self._connection.close()

        self._connection = None

    def producer(self, task_id):
        self.establish_connection()

        return BackendProducer(channel=self._connection, routing_key=task_id)

    def consumer(self, task_id, task_callback):
        self.establish_connection()

        return BackendConsumer(routing_key=task_id, callbacks=task_callback, channel=self._connection)

    def _save_task(self, task_id, status=TaskStatusEnum.CREATED.value, result=None, traceback=None):
        if not task_id or status not in TaskStatusEnum:
            return False

        producer = self.producer(task_id=task_id)

        task_dict = {
            "task_id": task_id,
            "status": status,
            "result": result,
            "traceback": traceback,
        }
        # 默认发送初始化时候的路由队列
        producer.publish_task(task_dict)
        producer.close()
        return True

    def mark_task_success(self, task_id, result=None):
        if not task_id:
            return False

        return self._save_task(task_id=task_id, status=TaskStatusEnum.SUCCESS.value, result=result)

    def mark_task_fail(self, task_id, result=None):
        if not task_id:
            return False

        return self._save_task(task_id=task_id, status=TaskStatusEnum.FAIL.value, traceback=result)

    def mark_task_retry(self, task_id, result=None):
        if not task_id:
            return False

        return self._save_task(task_id=task_id, status=TaskStatusEnum.RETRY.value, traceback=result)

    def get_task_result(self, task_id, timeout=None):
        if not task_id:
            return False, None

        result = []  # 利用可变数据记录结果

        def process_task_callback(body, message):
            print("task_content %s" % body)
            message.ack()
            result.append(body)

        consumer = self.consumer(task_id, process_task_callback)
        sleep_time = 0

        while True:
            result = consumer.qos(prefetch_count=1)
            if result:
                break
            else:
                if timeout is None:
                    break

                if timeout and sleep_time >= timeout:
                    break

                time.sleep(1)
                sleep_time += 1

        consumer.close()

        if not result:
            return True, None

        result = result[0]
        status = result["status"]
        if status == TaskStatusEnum.FAIL.value:
            return True, result["traceback"]

        return True, result["result"]


class DbBackend(BaseBackend):
    """
    利用数据库实现消息持久化
    """
    def __init__(self, backend_url=settings.BACKEND_URL, timeout=None):
        super(DbBackend, self).__init__(backend_url=backend_url, timeout=timeout)
        self._session = None

    def establish_connection(self):
        if self._connection:
            return

        self._connection = create_engine(self.backend_url, pool_timeout=self.timeout)
        self._connection.connect()

    def close_connection(self):
        """
        关闭连接
        :return:
        """
        if not self._connection:
            return

        self._connection.close()
        self._session.close_all()
        self._connection = None
        self._session = None

    def get_session(self):
        if self._session:
            return self._session

        self.establish_connection()
        session = sessionmaker(bind=self._connection)
        self._session = session()
        return self._session

    def get_task(self, task_id):
        try:
            return self.get_session().query(Task).filter(Task.task_id == task_id).one()
        except (exc.MultipleResultsFound, exc.NoResultFound):
            return None

    def get_tasks_by_ids(self, task_id_list):
        if not task_id_list:
            return []

        return self.get_session().query(Task).filter(Task.task_id.in_([task_id_list])).all()

    def _save_task(self, task_id, status=TaskStatusEnum.CREATED.value, result=None, traceback=None):
        if not task_id or status not in TaskStatusEnum.all_values():
            return False

        if result is not None:
            result = pickle.dumps(result, protocol=pickle.HIGHEST_PROTOCOL)
        elif traceback is not None:
            traceback = pickle.dumps(traceback, protocol=pickle.HIGHEST_PROTOCOL)

        self.get_session().add(Task(task_id=task_id, status=status, _result=result, _traceback=traceback))
        self.get_session().commit()
        return True

    def mark_task_retry(self, task_id, result=None):
        if not task_id:
            return False

        return self._save_task(task_id=task_id, status=TaskStatusEnum.RETRY.value, result=result)

    def mark_task_success(self, task_id, result=None):
        if not task_id:
            return False

        return self._save_task(task_id=task_id, status=TaskStatusEnum.SUCCESS.value, result=result)

    def mark_task_fail(self, task_id, result=None):
        if not task_id:
            return False

        return self._save_task(task_id=task_id, status=TaskStatusEnum.FAIL.value, traceback=result)

    def _get_task_result(self, task_id):
        task = self.get_task(task_id)
        if not task:
            return False, None

        if task.status == TaskStatusEnum.FAIL.value:
            return True, task.traceback

        return True, task.result

    def get_task_result(self, task_id, timeout=None):
        if not task_id:
            return False, None

        sleep_time = 0
        while True:
            success, result = self._get_task_result(task_id)
            if success:
                return True, result

            if timeout is None:
                return True, None

            if timeout and sleep_time >= timeout:
                return True, None

            time.sleep(1)
            sleep_time += 1


class RedisBackend(BaseBackend):
    """
    使用redis进行持久化
    """
    def __init__(self, backend_url=settings.BACKEND_URL, timeout=None):
        super(RedisBackend, self).__init__(backend_url=backend_url, timeout=timeout)

    def establish_connection(self):
        if self._connection:
            return

        # todo: redis中不允许直接传URL，需要进行解析
        self._connection = Redis(host="localhost", socket_connect_timeout=self.timeout)

    def close_connection(self):
        """
        关闭连接
        :return:
        """
        if not self._connection:
            return

        self._connection = None

    def _save_task(self, task_id, status, result=None, traceback=None):
        if not task_id or status not in TaskStatusEnum.all_values():
            return False

        if result is not None:
            result = pickle.dumps(result, protocol=pickle.HIGHEST_PROTOCOL)
        elif traceback is not None:
            result = pickle.dumps(traceback, protocol=pickle.HIGHEST_PROTOCOL)

        self._connection.lpush(task_id, result)
        return True

    def mark_task_retry(self, task_id, result=None):
        if not task_id:
            return False

        return self._save_task(task_id=task_id, status=TaskStatusEnum.RETRY.value, result=result)

    def mark_task_success(self, task_id, result=None):
        """
        标记任务为已经完成
        :param task_id:
        :param result:
        :param trace:
        :return:
        """
        if not self._connection:
            self.establish_connection()

        return self._save_task(task_id, TaskStatusEnum.SUCCESS.value, result=result)

    def mark_task_fail(self, task_id, result=None):
        """
        标记任务为失败
        :param task_id:
        :param result:
        :param trace:
        :return:
        """
        if not self._connection:
            self.establish_connection()

        return self._save_task(task_id, TaskStatusEnum.FAIL.value, traceback=result)

    def get_task_result(self, task_id, timeout=None):
        """
        返回任务的结果
        :param task_id:
        :param timeout: 阻塞时间（None表示不阻塞直接拿，0表示无限阻塞，其余为阻塞秒数）
        :return:
        """
        if not self._connection:
            self.establish_connection()
        # 不用轮询的方式是因为在列表可能经常为空的情况下会导致多次请求
        result = self._connection.blpop([task_id], timeout=timeout)
        if result:
            return pickle.loads(result[1])
        return


TYPE_2_BACKEND_CLS = {
    'amqp': MQBackend,
    'database': DbBackend,
    'redis': RedisBackend,
}


def get_backend_by_settings(type="database"):
    """
    获取backend
    :param type:
    :return:
    """
    return TYPE_2_BACKEND_CLS.get(type, DbBackend)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-07-11
@file:      test_schedule.py
@contact:   georgewang1994@163.com
@desc:      测试schedule
"""
import datetime
from queue import Queue

from ammonia.tests.test_base import ammonia, TestDBBackendBase
from ammonia.worker.schedule import Schedule


@ammonia.task(eta=datetime.datetime.now())
def get_sum(a, b):
    return a + b


@ammonia.task(wait=datetime.timedelta(seconds=1))
def get_sum2(a, b):
    return a + b


class TestSchedule(TestDBBackendBase):
    """
    测试定时任务
    """
    def setUp(self):
        self.ready_queue = Queue()
        self.schedule = Schedule(ready_queue=self.ready_queue)
        super(TestSchedule, self).setUp()

    def test_schedule_for_eta(self):
        get_sum(1, 2)
        self.assertIsNotNone(get_sum.start_time)
        start_time = datetime.datetime.now()
        # 为了方便计算，重新对任务的开始时间进行了赋值
        get_sum.start_time = start_time
        self.schedule.time_task_list.append(get_sum)
        result, next_internal_time = self.schedule.get_need_execute_tasks(
            start_time - datetime.timedelta(seconds=0.5)
        )
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].task_name, get_sum.task_name)
        self.assertEqual(next_internal_time, 1000)

        result, next_internal_time = self.schedule.get_need_execute_tasks(
            start_time + datetime.timedelta(seconds=0.5)
        )
        self.assertEqual(len(result), 0)
        self.assertEqual(next_internal_time, 1000)

    def test_schedule_for_wait(self):
        get_sum2(1, 2)  # start_time成功计算赋值，设置为当前时间后的一秒
        self.assertIsNotNone(get_sum2.start_time)

        start_time = datetime.datetime.now()
        self.schedule.time_task_list.append(get_sum2)
        result, next_internal_time = self.schedule.get_need_execute_tasks(start_time)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].task_name, get_sum2.task_name)
        self.assertEqual(next_internal_time, 1000)

        result, next_internal_time = self.schedule.get_need_execute_tasks(
            start_time + datetime.timedelta(seconds=1)
        )
        self.assertEqual(len(result), 0)
        self.assertEqual(next_internal_time, 1000)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-06-30
@file:      discover.py
@contact:   georgewang1994@163.com
@desc:      ...
"""

import importlib
import os
from importlib import util

from ammonia.settings import BASE_PROJECT_PATH, TASK_MODULE_SUFFIX


class BaseLoader(object):

    @classmethod
    def find_tasks(cls, project_name):
        return NotImplementedError

    @classmethod
    def on_worker_start(cls, project_name):
        cls.find_tasks(project_name)


class Loader(BaseLoader):

    @classmethod
    def get_tasks_from_dictionary(cls, package_name):
        """
        获取到当前目录下的所有以_tasks为结尾的文件
        :param package_name
        :return:
        """
        if not util.find_spec(package_name, package="."):
            raise ModuleNotFoundError("在项目中未找到package[%s]，请检查是否是一级目录" % package_name)

        package_path = os.path.join(BASE_PROJECT_PATH, package_name)

        file_list = []
        for dir_path in os.listdir(package_path):
            if not os.path.isfile(os.path.join(package_path, dir_path)):
                continue

            basename = os.path.basename(dir_path)
            if not basename[:-3].endswith(TASK_MODULE_SUFFIX):
                continue

            file_list.append(basename)
        return file_list

    @classmethod
    def find_tasks(cls, project_name):
        """
        寻找项目下的任务
        :param project_name:
        :return:
        """
        task_list = cls.get_tasks_from_dictionary(project_name)
        if not task_list:
            print("在指定的项目【%s】中没有找到任务的任务模块" % project_name)
            return []

        mod_list = []
        for task in task_list:
            task_module = task[:-3]
            try:
                mod = importlib.import_module("%s.%s" % (project_name, task_module))
                mod_list.append(mod)
            except ModuleNotFoundError:
                print("无法从项目【%s】中导入任务模块【%s】" % (project_name, task_module))
                raise

        return mod_list



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-06-15
@file:      test_base.py
@contact:   georgewang1994@163.com
@desc:      测试用例基类
"""

import contextlib
from unittest import TestCase

import sqlalchemy.exc as sqlalchemy_exc
from sqlalchemy import create_engine
from sqlalchemy.orm import close_all_sessions
from sqlalchemy_utils import database_exists, create_database, drop_database

from ammonia import settings
from ammonia.app import Ammonia
from ammonia.backends.models import Base

ammonia = Ammonia(conf={"BACKEND_URL": settings.TEST_CASE_BACKEND_URL})


class TestDBBackendBase(TestCase):
    def setUp(self):
        with contextlib.suppress(sqlalchemy_exc.ProgrammingError):
            engine = create_engine(settings.TEST_CASE_BACKEND_URL, encoding='utf-8', echo=True)

            if database_exists(engine.url):
                drop_database(settings.TEST_CASE_BACKEND_URL)

            create_database(settings.TEST_CASE_BACKEND_URL)
            Base.metadata.create_all(engine)


    def tearDown(self):
        with contextlib.suppress(sqlalchemy_exc.ProgrammingError):
            engine = create_engine(settings.TEST_CASE_BACKEND_URL, encoding='utf-8', echo=True)

            close_all_sessions()
            if database_exists(engine.url):
                drop_database(settings.TEST_CASE_BACKEND_URL)



from setuptools import setup, find_packages

requirements = [
    'alembic==1.0.10',
    'amqp==2.4.2',
    'Click==7.0',
    'decorator==4.4.0',
    'itsdangerous==1.1.0',
    'kombu==4.5.0',
    'pbr==5.2.0',
    'six==1.12.0',
    'SQLAlchemy==1.3.3',
    'SQLAlchemy-Utils==0.34.0',
    'sqlparse==0.3.0',
    'Tempita==0.5.2',
    'vine==1.3.0',
    'PyMySQL==0.9.3',
    'redis==3.2.1',
]

setup(
    name="Ammonia",
    version="0.0.9",
    description="task queue",
    author="george wang",
    author_email="georgewang1994@163.com",
    url="https://github.com/GeorgeWang1994/Ammonia",
    install_requires=requirements,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    entry_points={
        'console_scripts': [
            'ammonia=ammonia.command.start:start',
        ],
    },
)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-06-08
@file:      test_task.py
@contact:   georgewang1994@163.com
@desc:      测试任务
"""

from ammonia import settings
from ammonia.base.registry import task_registry
from ammonia.state import TaskStatusEnum
from ammonia.tests.test_base import TestDBBackendBase, ammonia


@ammonia.task(routing_key=settings.LOW_TASK_ROUTING_KEY)
def test_basic_task_param_func(a, b):
    return a + b


@ammonia.task(retry=3)
def test_task_retry_func(a, b):
    return a + b


package1 = ammonia.create_package("test_create_package", dependent=True)


@ammonia.task(package=package1)
def test_task_package_task1(a, b):
    return a + 2, b + 3


@ammonia.task(package=package1)
def test_task_package_task2(a, b):
    return a * b


class TestTask(TestDBBackendBase):
    def setUp(self):
        super().setUp()

    def test_basic_task_param(self):
        """
        测试基础的任务的参数的获取
        :return:
        """
        # 直接调用
        result = test_basic_task_param_func(1, 2)
        self.assertEqual(result, 3)

        task = task_registry.get('test_task.test_basic_task_param_func')
        self.assertEqual(task.status, TaskStatusEnum.SUCCESS.value)

    def test_task_try(self):
        """
        测试任务重试
        :return:
        """
        result = test_task_retry_func(1, "2")
        self.assertEqual(result, None)

        task = task_registry.get('test_task.test_task_retry_func')
        self.assertEqual(task.status, TaskStatusEnum.RETRY.value)

    def test_task_package(self):
        """
        测试任务包
        :return:
        """
        result = package1((1, 2))  # (1 + 2) * (2 + 3)
        self.assertEqual(result, 15)

        task = task_registry.get('test_create_package')
        self.assertEqual(task.status, TaskStatusEnum.SUCCESS.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
@author:    george wang
@datetime:  2019-05-11
@file:      models.py
@contact:   georgewang1994@163.com
@desc:      表存储
"""

import datetime
import pickle

from sqlalchemy import Column, String, DATETIME, BLOB
from sqlalchemy.ext.declarative import declarative_base

from ammonia.state import TaskStatusEnum

# 创建对象的基类:
Base = declarative_base()


class Task(Base):
    __tablename__ = "task"

    task_id = Column(String(100), primary_key=True, comment=u"任务id")
    status = Column(String(50), comment=u"任务当前的状态", nullable=False, default=TaskStatusEnum.CREATED.value)
    _result = Column(BLOB, comment=u"任务执行的结果", default=b"", nullable=True)
    _traceback = Column(BLOB, comment=u"报错信息", default=b"", nullable=True)
    create_time = Column(DATETIME, comment=u"创建时间", default=datetime.datetime.now(), nullable=False)
    modified_time = Column(DATETIME, comment=u"修改时间", default=datetime.datetime.now(), nullable=False)

    def __repr__(self):
        return "task:%s|%s" % (self.task_id, self.status)

    @property
    def result(self):
        if not self._result:
            return ""

        try:
            return pickle.loads(self._result)
        except TypeError:
            return ""

    @result.setter
    def result(self, result=""):
        self._result = pickle.dumps(result, protocol=pickle.HIGHEST_PROTOCOL)

    @property
    def traceback(self):
        if not self._traceback:
            return ""

        try:
            return pickle.loads(self._traceback)
        except TypeError:
            return ""

    @traceback.setter
    def traceback(self, traceback=""):
        self._traceback = pickle.dumps(traceback, protocol=pickle.HIGHEST_PROTOCOL)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
@author:    george wang
@datetime:  2019-06-16
@file:      worker.py
@contact:   georgewang1994@163.com
@desc:      worker
"""

from argparse import ArgumentParser

from ammonia.base.loader import Loader
from ammonia.worker.controller import WorkerController

# 命令提示信息
USAGE_MSG = "python --worker 10(worker count) --project example(your_project_name)"


def parse_options(arguments):
    parser = ArgumentParser(usage=USAGE_MSG)

    parser.add_argument("-w", "--worker", dest="worker_num", default=3, type=int, help="运行worker的个数")
    parser.add_argument("-p", "--project", dest="project_name", default="", type=str, help="项目名称")

    args = parser.parse_args(args=arguments)
    return args


# 启动信息
START_UP_MSG = """
tasks: {tasks}
"""


class Worker(object):

    def __init__(self, worker_num=10, project_name=""):
        self.project_name = project_name
        self.worker_controller = WorkerController(pool_worker_count=worker_num)
        self.loader = Loader()

    def setup(self):
        result = self.loader.find_tasks(self.project_name)
        print(START_UP_MSG.format(tasks=result))

    def run(self):
        self.worker_controller.start()
